import numpy as np
from .main import *


def _define(name: str, word) -> None:
    globals()[name] = word

c = Constant()
timestamp = NullaryWord('timestamp', timestamp_stack_function)
c_range = ConstantRange()
pandas = Pandas()
csv = CSV()
add = get_binary_operator('add', np.add)
sub = get_binary_operator('sub', np.subtract)
mul = get_binary_operator('mul', np.multiply)
pow = get_binary_operator('pow', np.power)
div = get_binary_operator('div', np.divide)
mod = get_binary_operator('mod', np.mod)
eq = get_binary_operator('eq', np.equal)
ne = get_binary_operator('ne', np.not_equal)
ge = get_binary_operator('ge', np.greater_equal)
gt = get_binary_operator('gt', np.greater)
le = get_binary_operator('le', np.less_equal)
lt = get_binary_operator('lt', np.less)
neg = get_unary_operator('neg', np.negative)
inv = get_unary_operator('inv', np.reciprocal)
abs = get_unary_operator('abs', np.abs)
sqrt = get_unary_operator('sqrt', np.sqrt)
exp = get_unary_operator('exp', np.exp)
log = get_unary_operator('log', np.log)
zero_to_na = get_unary_operator('zero_to_na', zero_to_na_op)
is_na = get_unary_operator('is_na', is_na_op)
logical_not = get_unary_operator('logical_not', logical_not_op)
roll = NullaryWord('roll', roll_stack_function)
swap = NullaryWord('swap', swap_stack_function)
drop = NullaryWord('drop', drop_stack_function)
dup = NullaryWord('dup', dup_stack_function)
rev = NullaryWord('rev', rev_stack_function)
clear = NullaryWord('clear', clear_stack_function)
hsort = NullaryWord('hsort', hsort_stack_function)
interleave = Interleave()
pull = Pull()
hpull = HeaderPull()
hfilter = HeaderFilter()
ewma = EWMA()
call = Call()
partial = Partial()
each = Each()
repeat = Repeat()
heach = NullaryWord('heach',heach_stack_function)
cleave = Cleave()
hset = HeaderSet()
hformat = HeaderFormat()
happly = HeaderApply()
rolling = Rolling()
expanding = Expanding()
crossing = NullaryWord('crossing', crossing_op)
rev_expanding = NullaryWord('rev_expanding', rev_expanding_op)
fill = Fill()
ffill = FFill()
join = Join()
cumsum = NullaryWord('cumsum', cumsum_stack_function)
sum = get_window_operator('sum',np.nansum, sum_oned_op)
max = get_window_operator('max',np.nanmax,max_oned_op)
min = get_window_operator('min',np.nanmin,min_oned_op)
prod = get_window_operator('prod',np.nanprod,prod_oned_op)
mean = get_window_operator('mean',np.nanmean, expanding_mean)
var = get_window_operator('var', np.nanvar, expanding_var)
std = get_window_operator('std', np.nanstd, expanding_var)
change = get_window_operator('change', change_twod_op, change_oned_op)
pct_change = get_window_operator('pct_change', pct_change_twod_op, pct_change_oned_op)
log_change = get_window_operator('log_change', log_change_twod_op, log_change_oned_op)
first = get_window_operator('first', first_twod_op, first_oned_op)
last = get_window_operator('last',last_twod_op,last_oned_op)
saved = Saved('saved')
quote = Quotation()
q = Quotation()
begin = Word('')


zscore = quote(quote(std).quote(last).quote(mean).cleave(3, depth=1).sub.swap.div).call



from __future__ import annotations
import os
import redis
import struct
import warnings
import numpy as np
import pandas as pd
from dataclasses import dataclass
from collections import namedtuple
from typing import Union, Dict
from redis.connection import UnixDomainSocketConnection
from .dates import datelike
from .ranges import Range
from . import periodicities

def get_client() -> Db:
    global _CLIENT
    if not '_CLIENT' in globals():
        args = {}
        if 'PYNTO_REDIS_PASSWORD' in os.environ:
            args['password'] = os.environ['PYNTO_REDIS_PASSWORD']
        if 'PYNTO_REDIS_PATH' in os.environ:
            args['path'] = os.environ['PYNTO_REDIS_PATH']
        else:
            if 'PYNTO_REDIS_HOST' in os.environ:
                args['host'] = os.environ['PYNTO_REDIS_HOST']
            if 'PYNTO_REDIS_PORT' in os.environ:
                args['port'] = os.environ['PYNTO_REDIS_PORT']
        _CLIENT = Db(**args)
    return _CLIENT

Type = namedtuple('Type', ['code', 'pad_value', 'length'])

TYPES: Dict[str, Type]  = {
    '<f8': Type('<f8',np.nan,8),
    '|b1': Type('|b1',False,1),
    '<i8': Type('<i8',0,8)
}

METADATA_FORMAT = '<6s6sll'
METADATA_SIZE = struct.calcsize(METADATA_FORMAT)

@dataclass
class Metadata:
    dtype: str
    periodicity_code: str
    start: int
    stop: int

    @property
    def is_frame(self):
        return self.dtype.startswith('<U')

    @property
    def periodicity(self):
        return periodicities.from_pandas(self.periodicity_code)

class Db:

    def __init__(self, **kwargs):
        if 'path' in kwargs:
           kwargs['connection_class'] = UnixDomainSocketConnection
        self._pool = redis.ConnectionPool(**kwargs)

    @property
    def connection(self):
        return redis.Redis(connection_pool=self._pool)

    def __setitem__(self, key: str, pandas: Union[pd.Series, pd.DataFrame]):
        if not isinstance(pandas.index, pd.MultiIndex) and pandas.index.freq is None:
            raise Exception('Missing index freq.')
        series_to_save = []
        if isinstance(pandas, pd.Series):
            series_to_save.append((key, pandas))
        else:
            self._check_dups(pandas.columns)

            if isinstance(pandas.index, pd.MultiIndex):
                assert len(pandas.index.levels) == 2
                rows = pandas.loc[pandas.index[0][0]].index
                self._check_dups(rows, 'row')
                rows = np.char.add('$',rows.values.astype('U'))
                columns = np.char.add('$',pandas.columns.values.astype('U'))
                columns = np.char.add(np.repeat(rows,len(columns)), np.tile(columns,len(rows)))
                index=pandas.index.levels[0]
                pandas = pd.DataFrame(pandas.values.reshape(len(index), len(columns)),
                            index=index, columns=columns) 
            periodicity = periodicities.from_pandas(pandas.index.freq.name)
            end = periodicity.get_index(pandas.index[-1].date()) + 1
            try:
                md = self._read_metadata(key)
                if md.periodicity != periodicity:
                    raise Exception('Incompatible periodicity.')
                columns = set(self._get_data_range(key, 0, -1).decode().split('\t'))
                if end > md.stop:
                    self._update_end(key, end)
                first_save = False
            except KeyError:
                start = periodicity.get_index(pandas.index[0].date())
                md = Metadata('<U', pandas.index.freq.name, start, end)
                columns = set()
                first_save = True
            new_columns = []
            for column,series in pandas.iteritems():
                series_code = f'{key}:{column}'
                if not column in columns:
                    columns.add(column)
                    new_columns.append(column)
                series = series[series.first_valid_index():series.last_valid_index()]
                series_to_save.append((series_code, series))
            if first_save:
                self._write(key, md, '\t'.join(new_columns).encode()) 
            elif len(new_columns) > 0:
                self.connection.append(key, ('\t' + '\t'.join(new_columns)).encode()) 
        for key, series in series_to_save:
            periodicity = periodicities.from_pandas(series.index.freq.name)
            start = periodicity.get_index(series.index[0].date())
            series_md = Metadata(series.dtype.str, series.index.freq.name,
                                    start, start + len(series.index)) 
            try:
                saved_md = self._read_metadata(key)
                if saved_md.periodicity != periodicity:
                    raise Exception(f'Incompatible periodicity.')   
            except:
                saved_md = None
            data = series.values
            if saved_md is None or series_md.start < saved_md.start:
                self._write(key, series_md, data.tobytes())
                continue
            if series_md.start > saved_md.stop:
                pad = np.full(series_md.start - saved_md.stop, TYPES[saved_md.dtype].pad_value)
                data = np.hstack([pad, data])
                start = saved_md.stop 
            else:
                start = series_md.start
            start_offset = (start - saved_md.start) * np.dtype(saved_md.dtype).itemsize
            self._set_data_range(key, start_offset, data.tobytes())
            if series_md.stop > saved_md.stop:
                self._update_end(key, series_md.stop)

    def __delitem__(self, key: str):
        try:
            md = self._read_metadata(key)
        except KeyError:
            return
        if md.is_frame:
            for series_key in self.read_frame_series_keys(key):
                self.connection.delete(series_key)
        self.connection.delete(key)

    def __getitem__(self, key: str):
        return self.read(key)

    def read(self, key: str,
                start: Union[int,datelike] = None,
                stop: Union[int,datelike] = None,
                periodicity: Union[str, periodicities.Periodicity] = None,
                resample_method: str = 'last') -> Union[pd.Series, pd.DataFrame]:
        md = self._read_metadata(key)
        if isinstance(periodicity, str):
            periodicity = getattr(periodicities, periodicity)
        if not md.is_frame:
            data = self.read_series_data(key, start, stop, periodicity, md, resample_method)
            return pd.Series(data[3], index=Range(*data[:3]).to_index(), name=key)
        else:
            data = self._read_frame_data(key, start, stop, periodicity, md, resample_method)
            index = Range(*data[:3]).to_index()
            c = np.array(data[3])
            array = data[4]
            if c[0].startswith('$') and c[0].count('$') == 2:
                columns, rows = {}, {}
                for a,b in np.char.split(np.char.lstrip(c,'$'),'$'):
                    rows[a] = None
                    columns[b] = None
                index = pd.MultiIndex.from_product([index, rows])
                array = array.reshape(array.shape[0]*len(rows),len(columns))
            else:
                columns = c
            return pd.DataFrame(array, columns=columns, index=index)

    def read_frame_diagonal(self, key, start=None, end=None, periodicity=None, resample_method='last'):
        md = self._read_metadata(key)
        start = md.start if start is None else get_index(md.periodicity, start)
        end = md.stop if end is None else get_index(md.periodicity, end)
        periodicity = md.periodicity if periodicity is None else periodicity
        index = Range(start, end, periodicity).to_index()
        c = np.array(self.read_frame_headers(key))
        if not (c[0].startswith('$') and c[0].count('$') == 2):
            raise ValueError(f'Frame {key} is not two-dimensional')
        columns = []
        for a,b in np.char.split(np.char.lstrip(c,'$'),'$'):
            if a == b:
                columns.append(a)
        array = np.column_stack([self.read_series_data(f'{key}:${c}${c}', start, end, periodicity,
                    resample_method)[3] for c in columns])
        return pd.DataFrame(array, columns=columns, index=index)

    def read_range(self, key: str) -> Range:
        md = self._read_metadata(key)
        return Range(md.start, md.stop, md.periodicity)

    def read_series_data(self, key: str,
                            start: Union[int,datelike] = None,
                            stop: Union[int,datelike] = None,
                            periodicity: periodicities.Periodicity = None,
                            md: Metadata = None,
                            resample_method: str = 'last'):
        if md is None:
            md = self._read_metadata(key)
        needs_resample = periodicity is not None and periodicity != md.periodicity
        if needs_resample:
            if not start is None:
                start_index = md.periodicity.get_index(
                        periodicity.get_date(periodicity.get_index(start)))
            else:
                start_index = md.start
            if not stop is None:
                stop_index = md.periodicity.get_index(
                        periodicity.get_date(periodicity.get_index(stop))) 
            else:
                stop_index = md.stop
        else:
            if start is None:
                start_index = md.start
            else:
                start_index = md.periodicity.get_index(start)
            if stop is None:
                stop_index = md.stop
            else:
                stop_index = md.periodicity.get_index(stop)
        periodicity = periodicity if periodicity else md.periodicity
        if start_index < md.stop and stop_index >= md.start:
            itemsize = np.dtype(md.dtype).itemsize 
            selected_start = max(0, start_index - md.start)
            selected_end = min(stop_index, md.stop + 2) - md.start
            buff = self._get_data_range(key, selected_start * itemsize, selected_end * itemsize)
            data = np.frombuffer(buff, md.dtype)
            if len(data) != stop_index - start_index:
                output_start = max(0, md.start - start_index)
                output = np.full(stop_index - start_index,TYPES[md.dtype].pad_value)
                output[output_start:output_start+len(data)] = data
            else:
                output = data
        else:
            output = np.full(stop_index - start_index,TYPES[md.dtype].pad_value)
        if needs_resample:
            s = pd.Series(output, index=Range(
                            start_index,stop_index,md.periodicity).to_index(), name=key)
            s = getattr(s.resample(periodicity.pandas_offset_code),
                            resample_method)().reindex(Range(start, stop, periodicity).to_index())
            return (periodicity.get_index(s.index[0].date()),
                    periodicity.get_index(s.index[-1].date()),
                    periodicity, s.values)
        else:
            return (start_index, stop_index, md.periodicity, output)
        
    def read_frame_headers(self, key):
        return self._get_data_range(key, 0, -1).decode().split('\t')

    def read_frame_series_keys(self, key):
        return [f'{key}:{c}' for c in self.read_frame_headers(key)]

    def _read_metadata(self, key: str) -> Metadata:
        data = self.connection.getrange(key,0,METADATA_SIZE-1)
        if len(data) == 0:
            raise KeyError(f'No metadata for key {key}')
        s = struct.unpack(METADATA_FORMAT, data)
        return Metadata(s[0].decode().strip(), s[1].decode().strip(), s[2], s[3])

    def _read_frame_data(self, key: str,
                            start: Union[int,datelike] = None,
                            stop: Union[int,datelike] = None,
                            periodicity: periodicities.Periodicity = None,
                            md: Metadata = None,
                            resample_method: str = 'last'):
        if md is None:
            md = self._read_metadata(key)
        start = md.start if start is None else md.periodicity.get_index(start)
        stop = md.stop if stop is None else md.periodicity.get_index(stop)
        periodicity = md.periodicity if periodicity is None else periodicity
        columns = self.read_frame_headers(key)
        data = np.column_stack([self.read_series_data(f'{key}:{c}', start, stop, periodicity,
                    None, resample_method)[3] for c in columns])
        return (start, stop, periodicity, columns, data)


    def _update_end(self, key: str, stop: int):
        self.connection.setrange(key, METADATA_SIZE - struct.calcsize('<l'), struct.pack('<l',int(stop)))

    def _write(self, key: str, metadata: Metadata, data: np.ndarray):
        packed_md = struct.pack(METADATA_FORMAT,
                                '{0: <6}'.format(metadata.dtype).encode(),
                                '{0: <6}'.format(metadata.periodicity.pandas_offset_code).encode(),
                                metadata.start,
                                metadata.stop) 
        self.connection.set(key, packed_md + data)

    def _get_data_range(self, key: str, start: int, stop: int):
        stop = -1 if stop == -1 else METADATA_SIZE + stop - 1
        return self.connection.getrange(key, str(METADATA_SIZE + start), str(stop))
        
    def _set_data_range(self, key: str, start: int, data: np.ndarray):
        self.connection.setrange(key, str(METADATA_SIZE + start), data)

    def _check_dups(self, index: pd.Index, type_='column'):
        unq, unq_cnt = np.unique(index, return_counts=True)
        if len(unq) != len(index):
            dups = unq[unq_cnt > 1].tolist()
            raise ValueError(f'Duplicate {type_} name{"s: " + str(dups)  if len(dups) > 1 else ": " + str(dups[0])}')





from .periodicities import *
from .ranges import Range
from .vocabulary import *
from .vocabulary import _define
from .database import get_client

db = get_client()

def define(name: str, word) -> None:
    globals()[name] = word
    _define(name, word)



import calendar
import datetime
import pandas as pd
from typing import Union


datelike: type = Union[int,str,datetime.date,datetime.datetime,pd.Timestamp]

def now() -> datetime.date:
    d = datetime.datetime.utcnow()
    if d.weekday() < 5 and d.hour >= 22:
        return d + pd.tseries.offsets.BDay()
    else:
        return d.date()

def parse_date(date: datelike) -> datetime.date:
    if isinstance(date, pd._libs.tslibs.timestamps.Timestamp):
        return date.date()
    elif isinstance(date, datetime.date):
        return date       
    elif isinstance(date, datetime.datetime):
        return date.date()
    elif isinstance(date, str):
        return datetime.datetime.strptime(date, "%Y-%m-%d").date()
    else:
        raise TypeError(f'{type(date)} is not datelike')

def round_b(date: datelike) -> datetime.date:
    date = parse_date(date)
    weekday = date.weekday()
    if weekday < 5:
        return date
    else:
        return date + datetime.timedelta(days=7 - weekday)

def round_w_fri(date: datelike) -> datetime.date:
    date = parse_date(date)
    weekday = date.weekday()
    if weekday == 4:
        return date
    else:
        return date + datetime.timedelta(days=4 - weekday)

def round_w_tue(date: datelike) -> datetime.date:
    date = parse_date(date)
    weekday = date.weekday()
    if weekday == 1:
        return date
    elif weekday == 0:
        return date + datetime.timedelta(days=1)
    else:
        return date + datetime.timedelta(days=8 - weekday)

def round_bm(date: datelike) -> datetime.date:
    date = parse_date(date)
    next_month = date.replace(day=28) + datetime.timedelta(days=4) 
    last_day = next_month - datetime.timedelta(days=next_month.day)
    return last_day - datetime.timedelta(days=max(0,last_day.weekday() - 4))

def round_ba_dec(date: datelike) -> datetime.date:
    date = parse_date(date)
    date = datetime.date(date.year, 12, 31)
    return round_bm(date)

def round_bq_dec(date: datelike) -> datetime.date:
    date = parse_date(date)
    remainder = date.month % 3
    plusmonths = 0 if remainder == 0 else 2 // remainder
    month = date.month + plusmonths
    year = date.year
    if  month > 12:
        month -= 12
        year += 1
    date = datetime.date(year, month, 28)
    return round_bm(date)

def add_bm(date: datelike, months: int) -> datetime.date:
    month = date.month - 1 + months
    year = date.year + month // 12
    month = month % 12 + 1
    day = min(date.day, calendar.monthrange(year,month)[1])
    day = datetime.date(year, month, day)
    return round_bm(day)

def add_b(date: datetime.date, b: int) -> datetime.date:
    daysSinceMonday = date.weekday()
    prevMonday = date - datetime.timedelta(days=daysSinceMonday)
    weekendsInDistance = (b + daysSinceMonday) // 5
    return prevMonday + datetime.timedelta(days=b + weekendsInDistance * 2 + daysSinceMonday)

def count_b(d1: datetime.date, d2: datetime.date) -> int:
    daysSinceMonday1 = d1.weekday()
    prevMonday1 = d1 - datetime.timedelta(days=daysSinceMonday1)
    daysSinceMonday2 = d2.weekday()
    prevMonday2 = d2 - datetime.timedelta(days=daysSinceMonday2)
    days = (prevMonday2 - prevMonday1).days
    days -= days * 2 // 7
    return days - daysSinceMonday1 + daysSinceMonday2

def count_w(d1: datetime.date, d2: datetime.date) -> int:
        return (d2 - d1).days // 7,

def offset_w(d: datelike,i: int ) -> datetime.date:
        return d + datetime.timedelta(days=i * 7) 

def count_m(d1: datetime.date, d2: datetime.date) -> int:
        return (d2.year - d1.year) * 12 + d2.month - d1.month

def count_q(d1: datetime.date, d2: datetime.date) -> int:
        return ((d2.year - d1.year) * 12 + d2.month - d1.month) // 3

def offset_q(d: datelike,i: int ) -> datetime.date:
        return add_bm(d, i * 3)

def count_y(d1: datetime.date, d2: datetime.date) -> int:
        return d2.year - d1.year

def offset_y(d: datelike,i: int ) -> datetime.date:
        return add_bm(d, i * 12)



import numpy as np


def expanding_mean(x, axis=None):
    nan_mask = np.isnan(x)
    cumsum = np.add.accumulate(np.where(nan_mask, 0, x))
    count = np.add.accumulate(np.where(nan_mask, 0, x / x))
    return np.where(nan_mask, np.nan, cumsum) / count

def expanding_var(x, axis=None):
    nan_mask = np.isnan(x)
    cumsum = np.add.accumulate(np.where(nan_mask, 0, x))
    cumsumOfSquares = np.add.accumulate(np.where(nan_mask, 0, x * x))
    count = np.add.accumulate(np.where(nan_mask, 0, x / x))
    return (cumsumOfSquares - cumsum * cumsum / count) / (count - 1)

def expanding_std(x, axis=None):
    nan_mask = np.isnan(x)
    cumsum = np.add.accumulate(np.where(nan_mask, 0, x))
    cumsumOfSquares = np.add.accumulate(np.where(nan_mask, 0, x * x))
    count = np.add.accumulate(np.where(nan_mask, 0, x / x))
    return np.sqrt((cumsumOfSquares - cumsum * cumsum / count) / (count - 1))

def ewma_vectorized_safe(data, alpha, row_size=None, dtype=None, order='C', out=None):
    data = np.array(data, copy=False)

    if dtype is None:
        if data.dtype == np.float32:
            dtype = np.float32
        else:
            dtype = np.float
    else:
        dtype = np.dtype(dtype)

    row_size = int(row_size) if row_size is not None else _get_max_row_size(alpha, dtype)

    if data.size <= row_size:
        # The normal function can handle this input, use that
        return ewma_vectorized(data, alpha, dtype=dtype, order=order, out=out)

    if data.ndim > 1:
        # flatten input
        data = np.reshape(data, -1, order=order)

    if out is None:
        out = np.empty_like(data, dtype=dtype)
    else:
        assert out.shape == data.shape
        assert out.dtype == dtype

    row_n = int(data.size // row_size)  # the number of rows to use
    trailing_n = int(data.size % row_size)  # the amount of data leftover
    first_offset = data[0]

    if trailing_n > 0:
        # set temporary results to slice view of out parameter
        out_main_view = np.reshape(out[:-trailing_n], (row_n, row_size))
        data_main_view = np.reshape(data[:-trailing_n], (row_n, row_size))
    else:
        out_main_view = out
        data_main_view = data

    # get all the scaled cumulative sums with 0 offset
    ewma_vectorized_2d(data_main_view, alpha, axis=1, offset=0, dtype=dtype,
                       order='C', out=out_main_view)

    scaling_factors = (1 - alpha) ** np.arange(1, row_size + 1)
    last_scaling_factor = scaling_factors[-1]

    # create offset array
    offsets = np.empty(out_main_view.shape[0], dtype=dtype)
    offsets[0] = first_offset
    # iteratively calculate offset for each row
    for i in range(1, out_main_view.shape[0]):
        offsets[i] = offsets[i - 1] * last_scaling_factor + out_main_view[i - 1, -1]

    # add the offsets to the result
    out_main_view += offsets[:, np.newaxis] * scaling_factors[np.newaxis, :]

    if trailing_n > 0:
        # process trailing data in the 2nd slice of the out parameter
        ewma_vectorized(data[-trailing_n:], alpha, offset=out_main_view[-1, -1],
                        dtype=dtype, order='C', out=out[-trailing_n:])
    return out

def _get_max_row_size(alpha, dtype=float):
    assert 0. <= alpha < 1.
    epsilon = np.finfo(dtype).tiny
    return int(np.log(epsilon)/np.log(1-alpha)) + 1

def ewma_vectorized(data, alpha, offset=None, dtype=None, order='C', out=None):
    data = np.array(data, copy=False)

    if dtype is None:
        if data.dtype == np.float32:
            dtype = np.float32
        else:
            dtype = np.float64
    else:
        dtype = np.dtype(dtype)

    if data.ndim > 1:
        # flatten input
        data = data.reshape(-1, order)

    if out is None:
        out = np.empty_like(data, dtype=dtype)
    else:
        assert out.shape == data.shape
        assert out.dtype == dtype

    if data.size < 1:
        # empty input, return empty array
        return out

    if offset is None:
        offset = data[0]

    alpha = np.array(alpha, copy=False).astype(dtype, copy=False)

    # scaling_factors -> 0 as len(data) gets large
    # this leads to divide-by-zeros below
    scaling_factors = np.power(1. - alpha, np.arange(data.size + 1, dtype=dtype),
                               dtype=dtype)
    # create cumulative sum array
    np.multiply(data, (alpha * scaling_factors[-2]) / scaling_factors[:-1],
                dtype=dtype, out=out)
    np.cumsum(out, dtype=dtype, out=out)

    # cumsums / scaling
    out /= scaling_factors[-2::-1]

    if offset != 0:
        offset = np.array(offset, copy=False).astype(dtype, copy=False)
        # add offsets
        out += offset * scaling_factors[1:]

    return out

def ewma_vectorized_2d(data, alpha, axis=None, offset=None, dtype=None, order='C', out=None):
    data = np.array(data, copy=False)

    assert data.ndim <= 2

    if dtype is None:
        if data.dtype == np.float32:
            dtype = np.float32
        else:
            dtype = np.float64
    else:
        dtype = np.dtype(dtype)

    if out is None:
        out = np.empty_like(data, dtype=dtype)
    else:
        assert out.shape == data.shape
        assert out.dtype == dtype

    if data.size < 1:
        # empty input, return empty array
        return out

    if axis is None or data.ndim < 2:
        # use 1D version
        if isinstance(offset, np.ndarray):
            offset = offset[0]
        return ewma_vectorized(data, alpha, offset, dtype=dtype, order=order,
                               out=out)

    assert -data.ndim <= axis < data.ndim

    # create reshaped data views
    out_view = out
    if axis < 0:
        axis = data.ndim - int(axis)

    if axis == 0:
        # transpose data views so columns are treated as rows
        data = data.T
        out_view = out_view.T

    if offset is None:
        # use the first element of each row as the offset
        offset = np.copy(data[:, 0])
    elif np.size(offset) == 1:
        offset = np.reshape(offset, (1,))

    alpha = np.array(alpha, copy=False).astype(dtype, copy=False)

    # calculate the moving average
    row_size = data.shape[1]
    row_n = data.shape[0]
    scaling_factors = np.power(1. - alpha, np.arange(row_size + 1, dtype=dtype),
                               dtype=dtype)
    # create a scaled cumulative sum array
    np.multiply(
        data,
        np.multiply(alpha * scaling_factors[-2], np.ones((row_n, 1), dtype=dtype),
                    dtype=dtype)
        / scaling_factors[np.newaxis, :-1],
        dtype=dtype, out=out_view
    )
    np.cumsum(out_view, axis=1, dtype=dtype, out=out_view)
    out_view /= scaling_factors[np.newaxis, -2::-1]

    if not (np.size(offset) == 1 and offset == 0):
        offset = offset.astype(dtype, copy=False)
        # add the offsets to the scaled cumulative sums
        out_view += offset[:, np.newaxis] * scaling_factors[np.newaxis, 1:]

    return out


def corrcoeff_1d(A,B):
    # Rowwise mean of input arrays & subtract from input arrays themeselves
    A_mA = A - A.mean(-1,keepdims=1)
    B_mB = B - B.mean(-1,keepdims=1)
    
    # Sum of squares
    ssA = np.einsum('i,i->',A_mA, A_mA)
    ssB = np.einsum('i,i->',B_mB, B_mB)
    
    # Finally get corr coeff
    return np.einsum('i,i->',A_mA,B_mB)/np.sqrt(ssA*ssB)

# https://stackoverflow.com/a/40085052/ @ Divakar
def strided_app(a, L, S ):  # Window len = L, Stride len/stepsize = S
    nrows = ((a.size-L)//S)+1
    n = a.strides[0]
    return np.lib.stride_tricks.as_strided(a, shape=(nrows,L), strides=(S*n,n))

# https://stackoverflow.com/a/41703623/ @Divakar
def corr2_coeff_rowwise(A,B):
    # Rowwise mean of input arrays & subtract from input arrays themeselves
    A_mA = A - A.mean(-1,keepdims=1)
    B_mB = B - B.mean(-1,keepdims=1)

    # Sum of squares across rows
    ssA = np.einsum('ij,ij->i',A_mA, A_mA)
    ssB = np.einsum('ij,ij->i',B_mB, B_mB)

    # Finally get corr coeff
    return np.einsum('ij,ij->i',A_mA,B_mB)/np.sqrt(ssA*ssB)
def nancorrcoeff_1d(A,B):
    # Get combined mask
    comb_mask = ~(np.isnan(A) & ~np.isnan(B))
    count = comb_mask.sum()

    # Rowwise mean of input arrays & subtract from input arrays themeselves
    A_mA = A - np.nansum(A * comb_mask,-1,keepdims=1)/count
    B_mB = B - np.nansum(B * comb_mask,-1,keepdims=1)/count

    # Replace NaNs with zeros, so that later summations could be computed    
    A_mA[~comb_mask] = 0
    B_mB[~comb_mask] = 0

    ssA = np.inner(A_mA,A_mA)
    ssB = np.inner(B_mB,B_mB)

    # Finally get corr coeff
    return np.inner(A_mA,B_mB)/np.sqrt(ssA*ssB)

def nancorrcoeff_rowwise(A,B):
    # Input : Two 2D arrays of same shapes (mxn). Output : One 1D array  (m,)
    # Get combined mask
    comb_mask = ~(np.isnan(A) & ~np.isnan(B))
    count = comb_mask.sum(axis=-1,keepdims=1)

    # Rowwise mean of input arrays & subtract from input arrays themeselves
    A_mA = A - np.nansum(A * comb_mask,-1,keepdims=1)/count
    B_mB = B - np.nansum(B * comb_mask,-1,keepdims=1)/count

    # Replace NaNs with zeros, so that later summations could be computed    
    A_mA[~comb_mask] = 0
    B_mB[~comb_mask] = 0

    # Sum of squares across rows
    ssA = np.einsum('ij,ij->i',A_mA, A_mA)
    ssB = np.einsum('ij,ij->i',B_mB, B_mB)

    # Finally get corr coeff
    return np.einsum('ij,ij->i',A_mA,B_mB)/np.sqrt(ssA*ssB)




from __future__ import annotations
import re
import copy
import warnings
import numbers
import numpy as np
import numpy.ma as ma
import pandas as pd
import traceback
from collections import namedtuple
from dataclasses import dataclass, field
from typing import Callable, List, Dict, Any
from .ranges import Range
from .tools import *
from .database import get_client
from . import vocabulary
from . import periodicities

#import psutil
from multiprocessing import Pool
import multiprocessing.pool

_MULTIPROCESSING = True
_CACHING = True

def disable_multiprocessing(disable: bool = True):
    global _MULTIPROCESSING
    _MULTIPROCESSING = not disable

class NoDaemonProcess(multiprocessing.Process):
    @property
    def daemon(self):
        return False

    @daemon.setter
    def daemon(self, value):
        pass

class NoDaemonContext(type(multiprocessing.get_context())):
    Process = NoDaemonProcess

class NestablePool(multiprocessing.pool.Pool):
    def __init__(self, *args, **kwargs):
        kwargs['context'] = NoDaemonContext()
        super(NestablePool, self).__init__(*args, **kwargs)

def get_rows(col_and_range):
    return col_and_range[0].rows(col_and_range[1])

@dataclass
class Column:
    header : str
    trace: str
    rows_function: Callable[[Range,dict], np.ndarray]
    args: dict = field(default_factory=list)
    copies: int = 1
    _rows_cache : dict = field(default_factory=dict)

    def rows(self, range_: Range) -> np.ndarray:
        try:
            if not _CACHING:
                return self.rows_function(range_, self.args)
            key = (range_.start, range_.stop, str(range_.periodicity))
            if key in self._rows_cache:
                rows = self._rows_cache[key]
            else: 
                rows = self.rows_function(range_, self.args)
                if self.copies > 1:
                    self._rows_cache[key] = rows
            if self.copies == 1:
                self._rows_cache.clear()
            else:
                self.copies -= 1
            return rows
        except Exception as e:
            print('Error in {self.header}')
            raise e
            raise ValueError(f'{e} in rows for column "{self.header}" ({self.trace})') from e
            
@dataclass
class Word:
    name: str
    next_: Word = None
    prev: Word = None
    quoted: Word = None
    args: dict = None

    def __getattr__(self, name):
        return self.__add__(getattr(vocabulary,name))

    def __add__(self, other: Word) -> Word:
        this = self._copy()
        other = other._copy()
        other = other._head()[0]
        this.next_ = other
        other.prev = this
        return other._tail()

    def __getitem__(self, key) -> pd.DataFrame:
        range_ = key if isinstance(key, Range) else Range.from_indexer(key)
        return self._evaluate([], range_)

    def __str__(self):
        if self.quoted:
            return f'quote({self.quoted.__repr__()})'
        else:
            s = self.name
            if self.args:
                s += '(' + ', '.join([f'{k}={str(v)[:2000]}' for k,v in self.args.items() if k != self]) + ')'
        return s

    def __repr__(self, no_recurse=False):
        s = ''
        current = self
        while True:
            s = str(current) + s
            if current.prev is None:
                break
            else:
                s = '.' + s
                current = current.prev
        return s

    def __len__(self):
        return self._head()[1]

    def __call__(self, args = {}) -> Word:
        this = self._copy()
        if '__class__' in args:
            del(args['__class__'])
        if 'self' in args:
            del(args['self'])
        this.args = args
        return this

    def _operation(self, stack, args):
        pass

    @property
    def columns(self):
        return self._evaluate([], None)
    
    def quote(self, quoted: Word):
        self.next_ = Quotation(quoted=quoted)
        self.next_.prev = self
        return self.next_
    
    def _copy(self) -> Word:
        cls = self.__class__
        copied = cls.__new__(cls)
        copied.__dict__.update(self.__dict__)
        first = copied
        while copied.prev is not None:
            cls = copied.prev.__class__
            prev = cls.__new__(cls)
            prev.__dict__.update(copied.prev.__dict__)
            copied.prev = prev
            prev.next_ = copied
            copied = prev
        return first

    def _evaluate(self, stack: List[Column], range_: Range = None):
        assert not self.quoted, 'Cannot evaluate quotation.'
        current = self._head()[0]
        while True:
            if current.quoted:
                stack.append(Column('quotation','quotation', current.quoted))
            else:
                try:
                    if not current.args:
                        current = current()
                    current._operation(stack, current.args)
                except Exception as e:
                    raise e
                    traceback.print_exc()
                    error_msg = f' in word "{current.name}"'
                    if current.prev is not None:
                        error_msg += ' preceded by "'
                        prev_expr = repr(current.prev)
                        if len(prev_expr) > 150:
                            error_msg += '...'
                        error_msg += prev_expr[-150:] + '"'
                    import sys
                    raise type(e)((str(e) + error_msg).replace("'",'"')).with_traceback(sys.exc_info()[2])

                    raise SyntaxError(repr(e) + error_msg) from e
            if current.next_ is None:
                break
            current = current.next_
        if not range_ is None:
            if len(stack) == 0:
                return None
            if _MULTIPROCESSING:
                pool = NestablePool(multiprocessing.cpu_count()-1)
                #Pool(psutil.cpu_count(logical=False)-1)
                values = np.column_stack(pool.map(get_rows, [(col, range_) for col in stack]))
            else:
                values = np.column_stack([col.rows(range_) for col in stack])
            return pd.DataFrame(values,
                        columns=[col.header for col in stack], index=range_.to_index())
        else:
            return [col.header for col in stack]

    def _head(self) -> Word:
        count = 1
        current = self
        while current.prev is not None:
            count += 1
            current = current.prev
        return (current, count)

    def _tail(self) -> Word:
        current = self
        while current.next_ is not None:
            current = current.next_
        return current

@dataclass
class Quotation(Word):
    name: str = 'quotation'
    
    def __call__(self, quoted):
        self.quoted=quoted
        return super().__call__(locals())

@dataclass
class NullaryWord(Word):
    def __init__(self,
            name,
            stack_function: Callable[[List[Column],Dict[str,Any]],None],
            stack_function_args: Dict[str, Any] = {}):
        self.stack_function = stack_function
        self.stack_function_args = stack_function_args
        self.stack_function_args['name'] = name
        super().__init__(name)

    def __call__(self):
        return self

    def _operation(self, stack, args):
        self.stack_function(stack, self.stack_function_args)

    def __repr__(self):
        return super().__repr__()

def dummy_stack_function(stack, args):
    pass

def const_col(range_, args):
    return np.full(len(range_), args['value'])

@dataclass
class Constant(Word):
    name: str = 'append'

    def __call__(self, *values):
        return super().__call__(locals())

    def _operation(self, stack, args):
        for value in args['values']:
            stack.append(Column('constant', str(value), const_col, {'value': value}))

def timestamp_col(range_, args):
    return np.array(range_.to_index().astype('int'))

def timestamp_stack_function(stack, args):
    stack.append(Column('timestamp','timestamp', timestamp_col))


@dataclass
class ConstantRange(Word):
    name: str = 'c_range'

    def __call__(self, value):
        return super().__call__(locals())

    def _operation(self, stack, args):
        for value in range(args['value']):
            stack.append(Column('constant', str(self.args['value']), const_col, {'value': value}))

def frame_col(range_, args):
    col = args['col']
    if col.index.freq is None:
        try:
            col.index.freq = pd.infer_freq(col.index)
        except:
            raise Exception('Unable to determine periodicity of pandas data.')
    periodicity = periodicities.from_pandas(col.index.freq.name)
    if periodicity != range_.periodicity:
        col =  col.resample(range_.to_index()).asfreq()
    start = range_.periodicity.get_index(col.index[0].date())
    end = range_.periodicity.get_index(col.index[-1].date()) + 1
    values = []
    if range_.start < start:
        values.append(np.full(min(start - range_.start, len(range_)), np.nan))
    if range_.start < end and range_.stop >= start:
        values.append(col.values[max(range_.start - start,0):min(range_.stop - start, len(col))])
    if len(values) > 0:
        values = np.concatenate(values)
    else:
        values = pd.Series()
    if len(values) < len(range_):
        values = np.concatenate([values, np.full(len(range_) - len(values), np.nan)])
    return values

def frame_to_columns(stack, frame):
    if frame.index.freq is None:
        frame.index.freq = pd.infer_freq(frame.index)
    for header, col in frame.iteritems():
        stack.append(Column(header,f'csv {header}', frame_col, {'col': col}))

@dataclass
class Pandas(Word):
    name: str = 'pandas'

    def __call__(self, frame_or_series):
        return super().__call__(locals())

    def _operation(self, stack, args):
        frame = args['frame_or_series']
        if isinstance(frame, pd.core.series.Series):
            frame = frame.toframe()
        frame_to_columns(stack, frame)

@dataclass
class CSV(Word):
    name: str = 'csv'

    def __call__(self, csv_file, index_col=0, header='infer'):
        return super().__call__(locals())

    def _operation(self, stack, args):
        frame = pd.read_csv(args['csv_file'], index_col=args['index_col'],
                                    header=args['header'], parse_dates=True)
        _frame_to_columns(stack, frame)

def binary_operator_col(range_, args):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        values1 = args['col1'].rows(range_)
        values2 = args['col2'].rows(range_)
        return np.where(np.logical_or(np.isnan(values1), np.isnan(values2)), np.nan,
                    args['op'](values1,values2))

def binary_operator_stack_function(stack, args):
    col2 = stack.pop()
    col1 = stack.pop()
    stack.append(Column(col1.header,f'{col1.trace}.{col2.trace}.{args["name"]}',
                    binary_operator_col,
                    {'col1': col1, 'col2': col2, 'op': args['op']}))

def get_binary_operator(name, op):
    return NullaryWord(name, binary_operator_stack_function, {'op': op})

def unary_operator_col(range_, args):
    return args['op'](args['col'].rows(range_))

def unary_operator_stack_function(stack, args):
    col = stack.pop()
    stack.append(Column(col.header,f'{col.trace}.{args["name"]}',unary_operator_col,
                            {'op': args['op'], 'col': col}))

def get_unary_operator(name, op):
    return NullaryWord(name, unary_operator_stack_function, {'op': op})

def zero_to_na_op(x):
    return np.where(np.equal(x,0),np.nan,x)
def is_na_op(x):
    return np.where(np.isnan(x), 1, 0)

def logical_not_op(x):
    return np.where(np.logical_not(x), 1, 0)

# Stack manipulation
def dup_stack_function(stack, args):
    stack.append(stack[-1])

def roll_stack_function(stack, args):
    stack.insert(0,stack.pop())

def swap_stack_function(stack, args):
    stack.insert(-1,stack.pop())

def drop_stack_function(stack, args):
    stack.pop()

def rev_stack_function(stack, args):
    stack.reverse()

def clear_stack_function(stack, args):
    stack.clear()

def hsort_stack_function(stack, args):
    stack.sort(key=lambda c: c.header)

@dataclass
class Interleave(Word):
    name: str = 'interleave'

    def __call__(self, count=None, split_into=2): return super().__call__(locals())

    def _operation(self, stack, args):
        count = args['count'] if args['count'] else len(stack) // args['split_into']
        last = 0
        lists = []
        for i in range(len(stack)+1):
            if i % count == 0 and i != 0:
                lists.append(stack[i-count:i])
                last = i
        del(stack[:last])
        stack += [val for tup in zip(*lists) for val in tup]
@dataclass
class Pull(Word):
    name: str = 'pull'

    def __call__(self, start, end=None, clear=False): return super().__call__(locals())

    def _operation(self, stack, args):
        end = -args['start'] - 1 if args['end'] is None else -args['end']
        start = len(stack) if args['start'] == 0 else -args['start']
        pulled = stack[end:start]
        if args['clear']:
            del(stack[:])
        else:
            del(stack[end:start])
        stack += pulled

@dataclass
class HeaderPull(Word):
    name: str = 'hpull'

    def __call__(self, *headers, clear=False, exact_match=False):
        return super().__call__(locals())
    def _operation(self, stack, args):
        filtered_stack = []
        for header in args['headers']:
            to_del = []
            matcher = lambda c: header == c.header if args['exact_match'] else re.match(header,col.header) is not None
            for i,col in enumerate(stack):
                if matcher(col):
                    filtered_stack.append(stack[i])
                    to_del.append(i)
            to_del.sort(reverse=True)
            for i in to_del:
                del(stack[i])
        if args['clear']:
            del(stack[:])
        stack += filtered_stack

@dataclass
class HeaderFilter(HeaderPull):
    name: str = 'hfilter'
    def __call__(self, *headers, clear=True, exact_match=False):
        return super().__call__(*headers, clear=True, exact_match=False)

def ewma_col(range_, args):
    alpha = 2 /(args['window'] + 1.0)
    data = args['col'].rows(range_)
    idx = np.cumsum(np.where(~np.isnan(data),1,0)) - 1
    nans = np.where(~np.isnan(data),1,np.nan)
#     starting_nans = np.where(idx == -1,np.nan,1)
    data = data[~np.isnan(data)]
    if len(data) == 0:
        return np.full(idx.shape[0],np.nan)
    out = ewma_vectorized_safe(data,alpha)
    #return out[idx] * starting_nans
    return out[idx] * nans

@dataclass
class EWMA(Word):
    name: str = 'ewma'
    def __call__(self, window, fill_nans=True): return super().__call__(locals())
    def _operation(self, stack, args):
        col = stack.pop()
        window = args['window']
        stack.append(Column(col.header,f'{col.trace}.ewma({window})',
                ewma_col, {'window': window, 'col': col}))

# Combinators
@dataclass
class Call(Word):
    name: str = 'call'
    def __call__(self, depth=None, copy=False): return super().__call__(locals())
    def _operation(self, stack, args):
        assert stack[-1].header == 'quotation', 'call needs a quotation on top of stack'
        quoted = stack.pop().rows_function
        if 'depth' not in args or args['depth'] is None:
            depth = len(stack) 
        else:
            depth = args['depth']

        if depth != 0:
            this_stack = stack[-depth:]
            if 'copy' not in args or not args['copy']:
                del(stack[-depth:])
        else:
            this_stack = []
        quoted._evaluate(this_stack)
        stack.extend(this_stack)

def partial_stack_function(stack, args):
    stack.extend(args['stack'])

@dataclass
class Partial(Word):
    name: str = 'partial'
    def __call__(self, depth=1, copy=False): return super().__call__(locals())
    def _operation(self, stack, args):
        assert stack[-1].header == 'quotation', 'partial needs a quotation on top of stack'
        quoted = stack.pop()
        depth = args['depth']
        if depth != 0:
            this_stack = stack[-depth:]
            if not args['copy']:
                del(stack[-depth:])
            else:
                for col in this_stack:
                    col.copies += 1
        else:
            this_stack = []
        quoted.rows_function = NullaryWord('partial', partial_stack_function, {'stack': this_stack}) + quoted.rows_function
        stack.append(quoted)

@dataclass
class Each(Word):
    name: str = 'each'
    def __call__(self, start=0, end=None, every=1, copy=False): return super().__call__(locals())
    def _operation(self, stack, args):
        assert stack[-1].header == 'quotation'
        quote = stack.pop().rows_function
        end = 0 if args['end'] is None else -args['end']
        start = len(stack) if args['start'] == 0 else -args['start']
        selected = stack[end:start]
        assert len(selected) % args['every'] == 0, 'Stack not evenly divisible by every'
        if not args['copy']:
            del(stack[end:start])
        else:
            for col in selected:
                col.copies += 1
        for t in zip(*[iter(selected)]*args['every']):
            this_stack = list(t)
            quote._evaluate(this_stack)
            stack += this_stack
@dataclass
class Repeat(Word):
    name: str = 'repeat'
    def __call__(self, times): return super().__call__(locals())
    def _operation(self, stack, args):
        assert stack[-1].header == 'quotation'
        quote = stack.pop().rows_function
        for _ in range(args['times']):
            quote._evaluate(stack)

def heach_stack_function(stack, args):
    assert stack[-1].header == 'quotation'
    quote = stack.pop().rows_function
    new_stack = []
    for header in set([c.header for c in stack]):
        to_del, filtered_stack = [], []
        for i,col in enumerate(stack):
            if header == col.header:
                filtered_stack.append(stack[i])
                to_del.append(i)
        quote._evaluate(filtered_stack)
        new_stack += filtered_stack
        to_del.sort(reverse=True)
        for i in to_del:
            del(stack[i])
    del(stack[:])
    stack += new_stack

@dataclass
class Cleave(Word):
    name: str = 'cleave'
    def __call__(self, num_quotations=-1, depth=None, copy=False): return super().__call__(locals())
    def _operation(self, stack, args):
        if args['num_quotations'] < 0:
            args['num_quotations'] = len(stack)
        quotes = [quote.rows_function for quote in stack[-args['num_quotations']:]]
        del(stack[-args['num_quotations']:])
        depth = len(stack) if args['depth'] is None else args['depth']
        copied_stack = stack[-depth:] if depth != 0 else []
        if not args['copy'] and depth != 0:
            del(stack[-depth:])
        else:
            for col in copied_stack:
                col.copies += 1
        for quote in quotes:
            this_stack = copied_stack[:]
            quote._evaluate(this_stack)
            stack += this_stack


# Header
@dataclass
class HeaderSet(Word):
    name: str = 'hset'
    def __call__(self, *headers): return super().__call__(locals())
    def _operation(self, stack, args):
        headers = args['headers']
        if len(headers) == 1 and headers[0].find(',') != -1:
            headers = headers[0].split(',')
        start = len(stack) - len(headers)
        for i in range(start,len(stack)):
            stack[i] = Column(headers[i - start], stack[i].trace,
                                stack[i].rows_function, stack[i].args)

@dataclass
class HeaderFormat(Word):
    name: str = 'hformat'
    def __call__(self, format_string): return super().__call__(locals())
    def _operation(self, stack, args):
        col = stack.pop()
        stack.append(Column(args['format_string'].format(col.header),
                            col.trace, col.rows_function, col.args))

@dataclass
class HeaderApply(Word):
    name: str = 'happly'
    def __call__(self, header_function): return super().__call__(locals())
    def _operation(self, stack, args):
        col = stack.pop()
        stack.append(Column(args['header_function'](col.header), col.trace,
                            col.rows_function, col.args))

# Windows
def rolling_col(range_,args):
    col = args['col']
    window = args['window']
    periodicity = args['periodicity']
    exclude_nans = args['exclude_nans']
    lookback_multiplier = args['lookback_multiplier']
    if not exclude_nans:
        lookback_multiplier = 1
    lookback = (window - 1) * lookback_multiplier
    if periodicity is None or periodicity == str(range_.periodicity):
        resample = False
        expanded_range = copy.copy(range_)
    else:
        resample = True
        expanded_range = Range(range_.start_date(),
                                range_.end_date(), periodicity)
    expanded_range.start = expanded_range.start - lookback
    expanded = col.rows(expanded_range)
    if range_.stop is None:
        range_.stop = expanded_range.stop
    #length = len(expanded_range)
    #if expanded.shape[0] < length + lookback:
        #fill = np.full(length + lookback - expanded.shape[0], np.nan)
        #expanded = np.concatenate([fill,col.rows(expanded_range)])
    mask = ~np.isnan(expanded) if exclude_nans else np.full(expanded.shape,True)
    no_nans = expanded[mask]
    # Indexes of no_nan values in expanded
    indexes = np.add.accumulate(np.where(mask))[0]
    if no_nans.shape[-1] - window < 0:
        #warnings.warn('Insufficient non-nan values in lookback.')
        no_nans = np.hstack([np.full(window - len(no_nans),np.nan),no_nans])
    shape = no_nans.shape[:-1] + (no_nans.shape[-1] - window + 1, window)
    strides = no_nans.strides + (no_nans.strides[-1],)
    windows = np.lib.stride_tricks.as_strided(no_nans, shape=shape, strides=strides)
    td = np.full((expanded.shape[0],window), np.nan)
    td[indexes[-windows.shape[0]:],:] = windows
    if not resample:
        return td[lookback:]
    else:
        expanded_range.start = expanded_range.start + lookback
        return pd.DataFrame(td[lookback:],
                    index=expanded_range.to_index()).reindex(range_.to_index())
@dataclass
class Rolling(Word):
    name: str = 'rolling'
    def __call__(self, window=2, exclude_nans=True, periodicity=None, lookback_multiplier=2): return super().__call__(locals())
    def _operation(self, stack, args):
        col = stack.pop()
        col_args = args.copy()
        col_args.update({'col': col})
        stack.append(Column(col.header,f'{col.trace}.rolling({args["window"]})', rolling_col, col_args))

def expanding_col(range_,args):
    start_date=args['start_date']
    col = args['col']
    index = range_.periodicity.get_index(start_date)
    if range_.start is None:
        range_.start = index
        return col.rows(range_)
    expanded_range = copy.copy(range_)
    offset = expanded_range.start - index
    expanded_range.start = index
    values = col.rows(expanded_range).view(ma.MaskedArray)
    values[:offset] = ma.masked
    if range_.stop is None:
        range_.stop = expanded_range.stop
    return values

@dataclass
class Expanding(Word):
    name: str = 'expanding'
    def __call__(self, start_date): return super().__call__(locals())
    def _operation(self, stack, args):
        col = stack.pop()
        stack.append(Column(col.header,f'{col.trace}.expanding({args["start_date"]})',
            expanding_col, {'start_date': args['start_date'], 'col': col}))

def crossing_col(range_, args):
    if _MULTIPROCESSING:
        pool = NestablePool(multiprocessing.cpu_count()-1) #Pool(psutil.cpu_count(logical=False)-1)
        return np.column_stack(pool.map(get_rows, [(col, range_) for col in args['cols']]))
    else:
        return np.column_stack([col.rows(range_) for col in args['cols']])

def crossing_op(stack, stack_args):
    cols = stack[:]
    stack.clear()
    #headers = ','.join([str(col.header) for col in cols])
    stack.append(Column(cols[0].header, f'crossing',crossing_col, {'cols':  cols}))

def rev_expanding_col(range_, args):
    return args['col'].rows(range_)[::-1]

def rev_expanding_op(stack, args):
    col = stack.pop()
    stack.append(Column(col.header, f'{col.header}.rev_expanding',
                        rev_expanding_col, {'col': col}))

def window_operator_col(range_, args):
    values = args['col'].rows(range_)
    if len(values.shape) == 2:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            return np.where(np.all(np.isnan(values),axis=1), np.nan,
                        args['twod_operation'](values, axis=1))
    else:
        cum = args['oned_operation'](values, axis=None)
        if values.strides[0] < 0:
            return cum[::-1]
        if isinstance(values,ma.MaskedArray):
            return ma.array(cum,mask=values.mask).compressed()
        return cum 

def cumsum_col(range_, args):
    v = args['col'].rows(range_)
    nans = np.isnan(v)
    cumsum = np.nancumsum(v)
    v[nans] = -np.diff(np.concatenate([[0.],cumsum[nans]]))
    return np.cumsum(v)

def cumsum_stack_function(stack, args):
    col = stack.pop()
    stack.append(Column(col.header, f'{col.trace}.cumsum', cumsum_col, {'col': col}))

# Data cleaning
def fill_col(range_, args):
    x = args['col'].rows(range_).copy()
    x[np.isnan(x)] = args['value']
    return x

@dataclass
class Fill(Word):
    name: str = 'fill'
    def __call__(self, value): return super().__call__(locals())
    def _operation(self, stack, args):
        col = stack.pop()
        stack.append(Column(col.header,f'{col.trace}.fill(arg["value"])',
                        fill_col, {'col': col, 'value': args['value']}))

def ffill_col(range_, args):
    lookback = abs(args['lookback']) 
    if range_.start is None and lookback != 0:
        args['col'].rows(range_)
    expanded_range = copy.copy(range_)
    expanded_range.start -= lookback
    x = args['col'].rows(expanded_range)
    if range_.stop is None:
        range_.stop = expanded_range.stop
    idx = np.where(~np.isnan(x),np.arange(len(x)),0)
    np.maximum.accumulate(idx,out=idx)
    return x[idx][lookback:]

@dataclass
class FFill(Word):
    name: str = 'ffill'
    def __call__(self, lookback=0): return super().__call__(locals())
    def _operation(self, stack, args):
        col = stack.pop()
        stack.append(Column(col.header,f'{col.trace}.ffill',
                        ffill_col, {'col': col, 'lookback': args['lookback']}))

def join_col(range_, args):
    date = args['date']
    if not isinstance(date, int):
        date = range_.periodicity.get_index(date)
    if range_.stop is not None and range_.stop < date:
        return args['col1'].rows(range_)
    if range_.start and range_.start >= date:
        return args['col2'].rows(range_)
    r_first = copy.copy(range_)
    r_first.stop = date
    r_second = copy.copy(range_)
    r_second.start = date
    v_first = args['col1'].rows(r_first)
    v_second = args['col2'].rows(r_second)
    if range_.stop is None:
        range_.stop = r_second.stop
    if range_.start is None:
        range_.start = r_first.start
    return np.concatenate([v_first, v_second])

@dataclass
class Join(Word):
    name: str = 'join'
    def __call__(self, date): return super().__call__(locals())
    def _operation(self, stack, args):
        col2 = stack.pop()
        col1 = stack.pop()
        stack.append(Column('join',f'{col1.trace}.{col2.trace}.join({args["date"]})',
                                join_col, {'col1': col1, 'col2': col2, 'date': args['date']}))

def saved_col(range_, args):
    if range_.start is not None and range_.stop is not None and range_.periodicity is not None:
        return get_client().read_series_data(args['key'], range_.start,
                        range_.stop, range_.periodicity)[3]
    else:
        data = get_client().read_series_data(args['key'])
        values = data[3][range_.start: range_.stop: range_.periodicity]
        if range_.start is None:
            range_.start = data[0]
        elif range_.start < 0:
            range_.start = data[1] + range_.start
        else:
            range_.start = data[0] + range_.start

        if range_.stop is None:
            range_.stop = data[1]
        elif range_.stop < 0:
            range_.stop = data[1] + range_.stop
        else:
            range_.stop = data[0] + range_.stop
        range_.periodicity_code = data[2]
        return values

@dataclass
class Saved(Word):
    name: str = 'saved'
    def __call__(self, key): return super().__call__(locals())
    def _operation(self, stack, args):
        md = get_client()._read_metadata(args['key'])
        if not md.is_frame:
            stack.append(Column(args['key'], args['key'], saved_col, args))
        else:   
            for header in get_client().read_frame_headers(args['key']):
                col_args = args.copy()
                col_args.update({'key': f'{args["key"]}:{header}'})
                stack.append(Column(header, f'{args["key"]}:{header}', saved_col, col_args))

def window_operator_stack_function(stack, args):
    name = args['name']
    col = stack.pop()
    col_args = args.copy()
    col_args['col'] = col
    stack.append(Column(col.header, f'{col.trace}.{name}', window_operator_col, col_args))

def get_window_operator(name,  twod_operation, oned_operation):
    return NullaryWord(name,window_operator_stack_function,
                        {'twod_operation': twod_operation, 'oned_operation': oned_operation})

def sum_oned_op(x, axis):
    mask = np.isnan(x)
    return np.where(mask, np.nan, np.add.accumulate(np.where(mask,0,x)))

def max_oned_op(x, axis):
    mask = np.isnan(x)
    return np.where(mask, np.nan, np.maximum.accumulate(np.where(mask,0,x)))

def min_oned_op(x, axis):
    mask = np.isnan(x)
    return np.where(mask, np.nan, np.minimum.accumulate(np.where(mask,0,x)))

def prod_oned_op(x, axis):
    mask = np.isnan(x)
    return np.where(mask, np.nan, np.multiply.accumulate(np.where(mask,0,x)))

def change_twod_op(x, axis):
    return x[:,-1] - x[:,0]

def change_oned_op(x, axis):
    return x - x[0]

def pct_change_twod_op(x, axis):
    return x[:,-1] / x[:,0] - 1

def pct_change_oned_op(x, axis):
    return x / x[0] - 1

def log_change_twod_op(x, axis):
    return np.log(x[:,-1] / x[:,0])

def log_change_oned_op(x, axis):
    return np.log( x / x[0])

def first_twod_op(x, axis):
    return  x[:,0]

def first_oned_op(x, axis):
    return np.full(x.shape,x[0])

def last_twod_op(x, axis):
    return  x[:,-1]

def last_oned_op(x, axis):
    return  x

def lag(number):
    return rolling(number+1).first




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pynto",
    version="0.3",
    author="Peter Graf",
    author_email="peter@pynto.tech",
    description="Data analysis using a concatenative paradigm",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/punkbrwstr/pynto",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords='data analysis quantitative tabular concatenative functional',
    python_requires='>=3.6',
    install_requires=['numpy','pandas','python-dateutil', 'timestamps'],
)



import datetime
from dataclasses import dataclass
from typing import Callable
from .dates import *

@dataclass
class Periodicity:
    code: str
    epoque: datetime.date
    id_: int
    annualization_factor: int
    pandas_offset_code: str
    round_function: Callable[[datelike], datetime.date]
    count_function: Callable[[datetime.date, datetime.date], int]
    offset_function: Callable[[datetime.date, int], datetime.date]

    def get_index(self, date: Union[int,datetime.date]):
        if isinstance(date, int):
            return date
        return self.count_function(self.epoque, self.round_function(date))

    def get_date(self, index):
        return self.offset_function(self.epoque, index)

    def offset(self, date, offset):
        return self.get_date(self.get_index(date) + offset)

    def __str__(self):
        return self.code

    def __repr__(self):
        return self.code



B =  Periodicity('B', datetime.date(1970,1,1), 0, 260, 'B', round_b, count_b, add_b)
W_F = Periodicity('W_F', datetime.date(1970,1,2), 1, 52, 'W-FRI',
        round_w_fri, count_w, offset_w) 
W_T = Periodicity('W_T', datetime.date(1969,12,30), 1, 52, 'W-TUE',
        round_w_tue, count_w, offset_w)
M = Periodicity('M', datetime.date(1970,1,30), 2, 12, 'BM',
        round_bm, count_m, add_bm)
Q = Periodicity('Q', datetime.date(1970,3,31), 3, 4, 'BQ-DEC',
        round_bq_dec, count_q, offset_q)
Y = Periodicity('Y', datetime.date(1970,12,31), 3, 1, 'BA-DEC',
        round_ba_dec, count_y, offset_y)

_pandas_map = {'B': B, 'W-TUE': W_T, 'W-FRI': W_F, 'BM': M, 'BQ-DEC': Q, 'BA-DEC': Y}
from_pandas = _pandas_map.get



from __future__ import annotations
import calendar
import datetime
from typing import Union
import pandas as pd
from .dates import *
from . import periodicities

class Range:

    def __init__(self, start: Union[datelike,int],
                        stop: Union[datelike,int] = None, 
                        periodicity: Union[periodicities.Periodicity, str] = None):
        if periodicity is None:
            periodicity = periodicities.B
        elif isinstance(periodicity, str):
            periodicity = getattr(periodicities, periodicity)
        if start is None:
            start = 0
        elif not isinstance(start, int):
            start = periodicity.get_index(start)
        if stop is None:
            stop = periodicity.get_date(periodicity.get_index(datetime.date.today()) + 1)
        if not isinstance(stop, int):
            stop = periodicity.get_index(stop)
        self.start = start
        self.stop = stop
        self.periodicity = periodicity

    @classmethod
    def from_indexer(cls, indexer: Union[slice,int,datelike]):
        if isinstance(indexer, slice):
            return cls(indexer.start, indexer.stop, indexer.step)
        else:
            if not isinstance(indexer, int):
                indexer = periodicities.B.get_index(indexer)
            return cls(indexer, indexer + 1)

    @classmethod
    def from_index(cls, date_range: pd.DatetimeIndex):
        range_ =  cls(date_range[0], date_range[-1],
                    periodicities.from_pandas(date_range.freq.name))
        range_.stop += 1
        return range_

    @classmethod
    def change_periodicity(cls, range_: Range,
                            periodicity: Union[str,periodicities.Periodicity]):
        if isinstance(periodicity, str):
            periodicity = getattr(periodicities, periodicity)
        return cls(range_[0], range_[-1], periodicity)

    def __iter__(self):
        i = 0
        while i + self.start < self.stop:
            yield self[i]
            i += 1

    def __getitem__(self, index):
        if index >= 0:
            if index >= self.stop - self.start:
                raise IndexError
            return self.periodicity.get_date(index + self.start)
        else:
            if abs(index) > self.stop - self.start + 1:
                raise IndexError
            return self.periodicity.get_date(index + self.stop + 1)

    def __len__(self):
        assert self.stop >= self.start, f'Negative range length {self.start}-{self.stop}'
        return self.stop - self.start

                
    def __repr__(self):
        if len(self) == 0:
            return '[]'
        return f'[{self[0].strftime("%Y-%m-%d")}:{self[-1].strftime("%Y-%m-%d")}:{str(self.periodicity)}]'

    def expand(self, by):   
        expanded = self.__class__(self.start, self.stop, self.periodicity)
        if by > 0:
            expanded.stop += by
        elif by < 0:
            expanded.start += by
        return expanded

    def offset(self, by):   
        return self.__class__(self.start + by, self.stop + by, self.periodicity)

    def to_index(self):
        return pd.date_range(self[0], self[-2], freq=self.periodicity.pandas_offset_code)
